import React, { useMemo, useState, useEffect } from 'react';
import {
  Button,
  DropdownItem,
  Icon,
  nilUUID,
  useAppDispatch,
  useAuthentication,
  useSessionCache,
  wsDeleteState,
} from '../..';
import { deleteSaveState, selectSaveState } from '../../data-access/store/sessionSlice';
import { clearQB } from '../../data-access/store/querybuilderSlice';
import { clearSchema } from '../../data-access/store/schemaSlice';
import { Popover, PopoverContent, PopoverTrigger } from '../../components/layout/Popover';
import { useHandleDatabase } from './useHandleDatabase';
import { wsGetInsights } from '../../data-access/broker/wsInsightSharing';
import { InsightDialog } from '@/lib/insight-sharing';
import { Tooltip, TooltipTrigger, TooltipContent, TooltipProvider } from '@/lib/components/tooltip';
import { SaveState } from 'ts-common';

type Props = {
  onClose: () => void;
  saveStates: { [id: string]: SaveState };
  changeActive: (val: 'add' | 'update') => void;
  setSelectedSaveState: (val: SaveState) => void;
};

type insightStats = {
  fired: number;
  total: number;
};

type DatabaseTableHeaderTypes = 'name' | 'protocol' | 'url' | 'alarms';

const DataTableHeader = ({
  name,
  orderBy,
  setOrderBy,
}: {
  name: DatabaseTableHeaderTypes;
  orderBy: [DatabaseTableHeaderTypes, 'asc' | 'desc'];
  setOrderBy: (val: [DatabaseTableHeaderTypes, 'asc' | 'desc']) => void;
}) => {
  return (
    <th
      scope="col"
      className="text-left hover:underline cursor-pointer"
      onClick={() => setOrderBy([name, orderBy[1] === 'desc' && orderBy[0] === name ? 'asc' : 'desc'])}
    >
      <span className="flex items-center gap-0">
        {name}
        {orderBy[0] !== name && <div className="min-w-6 min-h-4 " />}
        {orderBy[0] === name && orderBy[1] === 'asc' && <Icon component="icon-[ic--baseline-arrow-drop-up]" />}
        {orderBy[0] === name && orderBy[1] === 'desc' && <Icon component="icon-[ic--baseline-arrow-drop-down]" />}
      </span>
    </th>
  );
};

export async function fetchallAlarmsStats(saveStateIds: string[]): Promise<{ [stateId: string]: insightStats }> {
  const allInsights: { [stateId: string]: insightStats } = {};

  for (const saveStateId of saveStateIds) {
    await new Promise<void>(resolve => {
      wsGetInsights({ saveStateId: saveStateId }, (data, status) => {
        if (status === 'success' && Array.isArray(data)) {
          // !TODO: get status of each alarms

          const alerts = data.filter(entry => entry.type === 'alert');

          if (alerts.length > 0) {
            allInsights[saveStateId] = { fired: 1, total: alerts.length };
          } else {
            allInsights[saveStateId] = { fired: 0, total: 0 };
          }
        }
        resolve();
      });
    });
  }

  return allInsights;
}

export function Databases({ onClose, saveStates, changeActive, setSelectedSaveState }: Props) {
  const dispatch = useAppDispatch();
  const auth = useAuthentication();
  const session = useSessionCache();
  const databaseHandler = useHandleDatabase();
  const [orderBy, setOrderBy] = useState<[DatabaseTableHeaderTypes, 'asc' | 'desc']>(['name', 'desc']);
  const [sharing, setSharing] = useState<boolean>(false);
  const [monitoringOpen, setMonitoringOpen] = useState<boolean>(false);
  const [allAlarmsStats, setAllAlarmsStats] = useState<{ [stateId: string]: insightStats }>({});

  const orderedSaveStates = useMemo(
    () =>
      Object.keys(saveStates).sort((a, b) => {
        if (orderBy[0] === 'alarms') {
          // Sort by the number of fired alarms
          const dir = orderBy[1] === 'asc' ? 1 : -1;
          const alarmsA = allAlarmsStats[a]?.fired || 0;
          const alarmsB = allAlarmsStats[b]?.fired || 0;

          if (alarmsA < alarmsB) return -dir;
          if (alarmsA > alarmsB) return dir;
          return 0;
        }

        const dir = orderBy[1] === 'asc' ? 1 : -1;
        if (orderBy[0] === 'name') {
          if (saveStates[a].name.toLowerCase() <= saveStates[b].name.toLowerCase()) return dir;
          else return -dir;
        } else {
          if (
            saveStates[a].dbConnections?.[0][orderBy[0]].toLowerCase() <=
            saveStates[b].dbConnections?.[0][orderBy[0]].toLowerCase()
          )
            return dir;
          else return -dir;
        }
      }),
    [saveStates, orderBy],
  );

  useEffect(() => {
    fetchallAlarmsStats(orderedSaveStates).then(insightsStats => {
      setAllAlarmsStats(insightsStats);
    });
  }, [orderedSaveStates]);

  return (
    <TooltipProvider delayDuration={300}>
      <div className="flex flex-col gap-4">
        <div className="flex gap-x-4">
          <Button
            label="Database"
            variantType="primary"
            iconComponent="icon-[ic--baseline-add]"
            iconPosition="leading"
            onClick={() => changeActive('add')}
          />
          <Button
            label="Learn how to use"
            variant="ghost"
            iconComponent="icon-[mdi--play]"
            iconPosition="leading"
            onClick={() => window.open('https://graphpolaris.com/', '_blank')}
          />
        </div>
        {orderedSaveStates.length > 0 && (
          <table className="w-full">
            <thead>
              <tr>
                <DataTableHeader name="name" orderBy={orderBy} setOrderBy={setOrderBy} />
                <DataTableHeader name="protocol" orderBy={orderBy} setOrderBy={setOrderBy} />
                <DataTableHeader name="url" orderBy={orderBy} setOrderBy={setOrderBy} />
                <DataTableHeader name="alarms" orderBy={orderBy} setOrderBy={setOrderBy} />
              </tr>
            </thead>
            <tbody className="divide-y divide-white overflow-auto table-fixed">
              {orderedSaveStates.map(key => (
                <tr key={key}>
                  <td>
                    <Button
                      variant={session.currentSaveState === key ? 'outline' : 'ghost'}
                      onClick={() => {
                        if (key !== session.currentSaveState) {
                          dispatch(selectSaveState(key));
                          dispatch(clearQB());
                          dispatch(clearSchema());
                          onClose();
                        }
                      }}
                    >
                      {saveStates[key].name}
                    </Button>
                  </td>
                  <td className="text-left">
                    <span className="font-light">{saveStates[key].dbConnections?.[0]?.protocol}</span>
                  </td>
                  <td className="text-left">
                    <span className="font-light">{saveStates[key].dbConnections?.[0]?.url}</span>
                  </td>
                  <td
                    className="text-left cursor-pointer group"
                    onClick={() => {
                      if (key !== session.currentSaveState) {
                        dispatch(selectSaveState(key));
                        dispatch(clearQB());
                        dispatch(clearSchema());
                        setMonitoringOpen(true);
                        onClose();
                      }
                    }}
                  >
                    {allAlarmsStats[key] && allAlarmsStats[key].total > 0 ? (
                      <Tooltip key={key + '-insight'}>
                        <TooltipTrigger asChild>
                          <div className="w-full h-4 flex items-center rounded border border-transparent group-hover:border-2 group-hover:border-secondary-900">
                            <div
                              className="h-full bg-danger-500 rounded-l-sm"
                              style={{
                                width: `${(allAlarmsStats[key].fired / allAlarmsStats[key].total) * 100}%`,
                              }}
                            />
                            <div
                              className="h-full bg-success-500 rounded-r-sm"
                              style={{
                                width: `${100 - (allAlarmsStats[key].fired / allAlarmsStats[key].total) * 100}%`,
                              }}
                            />
                          </div>
                        </TooltipTrigger>
                        <TooltipContent>
                          <span>{`Alarms: ${allAlarmsStats[key].fired} out of ${allAlarmsStats[key].total} fired`}</span>
                        </TooltipContent>
                      </Tooltip>
                    ) : (
                      <Tooltip key={key + '-alarms'}>
                        <TooltipTrigger asChild>
                          <span>No Alarms</span>
                        </TooltipTrigger>
                        <TooltipContent>
                          <span>No alarms data available for this entry</span>
                        </TooltipContent>
                      </Tooltip>
                    )}
                  </td>
                  <td className="text-right flex justify-end">
                    <Button
                      iconComponent="icon-[mdi--trash-outline]"
                      variant="ghost"
                      onClick={() => {
                        if (session.currentSaveState === key) {
                          dispatch(clearQB());
                          dispatch(clearSchema());
                        }
                        wsDeleteState({ saveStateID: key });
                        dispatch(deleteSaveState(key));
                      }}
                    />
                    <Popover>
                      <PopoverTrigger>
                        <Button iconComponent="icon-[mi--options-vertical]" variant="ghost" />
                      </PopoverTrigger>
                      <PopoverContent className="w-56 z-50 bg-light rounded-sm border-[1px] outline-none">
                        <DropdownItem
                          value="Open"
                          onClick={() => {
                            if (key !== session.currentSaveState) {
                              dispatch(selectSaveState(key));
                              dispatch(clearQB());
                              dispatch(clearSchema());
                            }
                            onClose();
                          }}
                        />
                        <DropdownItem
                          value="Update"
                          onClick={() => {
                            changeActive('update');
                            setSelectedSaveState(saveStates[key]);
                          }}
                        />
                        <DropdownItem
                          value="Clone"
                          onClick={() => {
                            databaseHandler.submitDatabaseChange(
                              { ...saveStates[key], name: saveStates[key].name + ' (copy)', id: nilUUID },
                              'add',
                              true,
                              () => {},
                            );
                            setSelectedSaveState(saveStates[key]);
                          }}
                        />
                        <DropdownItem
                          value={sharing ? 'Creating Share Link' : 'Share'}
                          disabled={sharing}
                          onClick={async () => {
                            setSharing(true);
                            await auth.shareLink();
                            setSharing(false);
                          }}
                        />
                        <DropdownItem
                          value="Delete"
                          onClick={() => {
                            if (session.currentSaveState === key) {
                              dispatch(clearQB());
                              dispatch(clearSchema());
                            }
                            wsDeleteState({ saveStateID: key });
                            dispatch(deleteSaveState(key));
                          }}
                          className="text-danger"
                        />
                      </PopoverContent>
                    </Popover>
                  </td>
                </tr>
              ))}
            </tbody>
          </table>
        )}
        <InsightDialog open={monitoringOpen} onClose={() => setMonitoringOpen(false)} />
      </div>
    </TooltipProvider>
  );
}
